@@ -8,7 +8,7 @@ function pkg_fileinfo(id::PkgId)
         ret = @static if VERSION ≥ v"1.11.0-DEV.683" # https://github.com/JuliaLang/julia/pull/49866
             io = open(cachepath, "r")
             checksum = Base.isvalid_cache_header(io)
-            iszero(checksum) && (close(io); return badfile)
+            iszero(checksum) && (close(io); return nothing)
             provides, (_, includes_srcfiles_only, requires), required_modules, _... =
                 Base.parse_cache_header(io, cachepath)
             close(io)
@@ -18,7 +18,7 @@ function pkg_fileinfo(id::PkgId)
             Base.parse_cache_header(cachepath; srcfiles_only = true)
         end
         ret
-    catch err
+    catch
         return nothing
     end
     includes, _ = includes_requires
@@ -110,3 +110,19 @@ function modulefiles(mod::Module)
     included_files = filter(mf->mf.id == id, includes)
     return keypath(parentfile), [keypath(mf.filename) for mf in included_files]
 end
+
+function modulefiles_basestlibs(id)
+    ret = Revise.pkg_fileinfo(id)
+    cachefile, includes = ret === nothing ? (nothing, nothing) : ret[1:2]
+    # `cachefile` will be nothing for Base and stdlibs that *haven't* been moved out
+    cachefile === nothing && return Iterators.drop(Base._included_files, 1)  # stepping through sysimg.jl rebuilds Base, omit it
+    # stdlibs that are packages
+    mod = Base.loaded_modules[id]
+    return map(includes) do inc
+        submod = mod
+        for sm in inc.modpath
+            submod = getfield(submod, Symbol(sm))
+        end
+        return (submod, inc.filename)
+    end
+end
