@@ -1,20 +1,9 @@
 ## Analyzing lowered code
 
-function assign_this!(frame, value)
+function assign_this!(frame::Frame, @nospecialize value)
     frame.framedata.ssavalues[frame.pc] = value
 end
 
-# TODO Define abstract type MethodInfo end
-
-# This defines the API needed to store signatures using methods_by_execution!
-# This default version is simple and only used for testing purposes.
-# The "real" one is CodeTrackingMethodInfo in Revise.jl.
-const MethodInfo = IdDict{Type,LineNumberNode}
-add_signature!(methodinfo, @nospecialize(sig), ln) = push!(methodinfo, sig=>ln)
-push_expr!(methodinfo, mod::Module, ex::Expr) = methodinfo
-pop_expr!(methodinfo) = methodinfo
-add_includes!(methodinfo, mod::Module, filename) = methodinfo
-
 function is_some_include(@nospecialize(f))
     @assert !isa(f, Core.SSAValue) && !isa(f, JuliaInterpreter.SSAValue)
     if isa(f, GlobalRef)
@@ -44,29 +33,6 @@ function is_defaultctors(@nospecialize(f))
     return false
 end
 
-# This is not generally used, see `is_method_or_eval` instead
-function hastrackedexpr(@nospecialize(stmt), code)
-    haseval = false
-    if isa(stmt, Expr)
-        haseval = matches_eval(stmt)
-        if stmt.head === :call
-            f = stmt.args[1]
-            while isa(f, Core.SSAValue) || isa(f, JuliaInterpreter.SSAValue)
-                f = code[f.id]
-            end
-            callee_matches(f, Core, :_typebody!) && return true, haseval
-            callee_matches(f, Core, :_setsuper!) && return true, haseval
-            is_some_include(f) && return true, haseval
-        elseif stmt.head === :thunk
-            newcode = (stmt.args[1]::Core.CodeInfo).code
-            any(s->any(hastrackedexpr(s, newcode)), newcode) && return true, haseval
-        elseif stmt.head === :method
-            return true, haseval
-        end
-    end
-    return false, haseval
-end
-
 function matches_eval(stmt::Expr)
     stmt.head === :call || return false
     f = stmt.args[1]
@@ -108,16 +74,15 @@ function defines_function(@nospecialize(ci))
 end
 
 """
-    isrequired, evalassign = minimal_evaluation!([predicate,] methodinfo, src::Core.CodeInfo, mode::Symbol)
+    isrequired, evalassign = minimal_evaluation!([predicate,] src::Core.CodeInfo, mode::Symbol)
 
 Mark required statements in `src`: `isrequired[i]` is `true` if `src.code[i]` should be evaluated.
-Statements are analyzed by `isreq, haseval = predicate(stmt)`, and `predicate` defaults
-to `Revise.is_method_or_eval`.
+Statements are analyzed by `isreq, haseval = predicate(stmt)`.
 `haseval` is true if the statement came from `@eval` or `eval(...)` call.
 Since the contents of such expression are difficult to analyze, it is generally
 safest to execute all such evals.
 """
-function minimal_evaluation!(@nospecialize(predicate), methodinfo, mod::Module, src::Core.CodeInfo, mode::Symbol)
+function minimal_evaluation!(@nospecialize(predicate), mod::Module, src::Core.CodeInfo, mode::Symbol)
     edges = CodeEdges(mod, src)
     # LoweredCodeUtils.print_with_code(stdout, src, edges)
     isrequired = fill(false, length(src.code))
@@ -173,15 +138,15 @@ function minimal_evaluation!(@nospecialize(predicate), methodinfo, mod::Module,
     # All tracked expressions are marked. Now add their dependencies.
     # LoweredCodeUtils.print_with_code(stdout, src, isrequired)
     lines_required!(isrequired, src, edges;)
-                    # norequire=mode===:sigs ? LoweredCodeUtils.exclude_named_typedefs(src, edges) : ())
+                    # norequire = mode === :sigs ? LoweredCodeUtils.exclude_named_typedefs(src, edges) : ())
     # LoweredCodeUtils.print_with_code(stdout, src, isrequired)
     return isrequired, evalassign
 end
-@noinline minimal_evaluation!(@nospecialize(predicate), methodinfo, frame::Frame, mode::Symbol) =
-    minimal_evaluation!(predicate, methodinfo, moduleof(frame), frame.framecode.src, mode)
+@noinline minimal_evaluation!(@nospecialize(predicate), frame::Frame, mode::Symbol) =
+    minimal_evaluation!(predicate, moduleof(frame), frame.framecode.src, mode)
 
-function minimal_evaluation!(methodinfo, frame::Frame, mode::Symbol)
-    minimal_evaluation!(methodinfo, frame, mode) do @nospecialize(stmt), code::Vector{Any}
+function minimal_evaluation!(frame::Frame, mode::Symbol)
+    minimal_evaluation!(frame, mode) do @nospecialize(stmt), code::Vector{Any}
         ismeth, haseval, isinclude, isnamespace, istoplevel = categorize_stmt(stmt, code)
         isreq = ismeth | isinclude | istoplevel
         return mode === :sigs ? (isreq, haseval) : (isreq | isnamespace, haseval)
@@ -189,27 +154,29 @@ function minimal_evaluation!(methodinfo, frame::Frame, mode::Symbol)
 end
 
 function methods_by_execution(mod::Module, ex::Expr; kwargs...)
-    methodinfo = MethodInfo()
-    value, thk = methods_by_execution!(JuliaInterpreter.Compiled(), methodinfo, mod, ex; kwargs...)
-    return methodinfo, thk
+    exinfo = ExInfo(ex)
+    _, thk = methods_by_execution!(JuliaInterpreter.Compiled(), exinfo, mod, ex; kwargs...)
+    return exinfo, thk
 end
 
 """
-    methods_by_execution!([interp::Interpreter=JuliaInterpreter.Compiled(),] methodinfo, mod::Module, ex::Expr;
-                          mode=:eval, disablebp=true, skip_include=mode!==:eval, always_rethrow=false)
+    methods_by_execution!(
+        [interp::Interpreter=JuliaInterpreter.Compiled(),] exinfo::ExInfo, mod::Module, ex::Expr;
+        mode::Symbol = :eval, disablebp::Bool = true, skip_include::Bool = mode!==:eval, always_rethrow::Bool = false
+    )
 
 Evaluate or analyze `ex` in the context of `mod`.
 Depending on the setting of `mode` (see the Extended help), it supports full evaluation or just the minimal
 evaluation needed to extract method signatures.
 `interp` controls JuliaInterpreter's evaluation of any non-intercepted statement;
 likely choices are `JuliaInterpreter.Compiled()` or `JuliaInterpreter.RecursiveInterpreter()`.
-`methodinfo` is a cache for storing information about any method definitions (see [`CodeTrackingMethodInfo`](@ref)).
+`exinfo` is a cache for storing information about any method definitions (see [`ExInfo`](@ref)).
 
 # Extended help
 
 The action depends on `mode`:
 
-- `:eval` evaluates the expression in `mod`, similar to `Core.eval(mod, ex)` except that `methodinfo`
+- `:eval` evaluates the expression in `mod`, similar to `Core.eval(mod, ex)` except that `exinfo`
   will be populated with information about any signatures. This mode is used to implement `includet`.
 - `:sigs` analyzes `ex` and extracts signatures of methods (specifically, statements flagged by
   [`Revise.minimal_evaluation!`](@ref)), but does not evaluate `ex` in the traditional sense.
@@ -231,14 +198,16 @@ The other keyword arguments are more straightforward:
 
 - `disablebp` controls whether JuliaInterpreter's breakpoints are disabled before stepping through the code.
   They are restored on exit.
-- `skip_include` prevents execution of `include` statements, instead inserting them into `methodinfo`'s
+- `skip_include` prevents execution of `include` statements, instead inserting them into `exinfo`'s
   cache. This defaults to `true` unless `mode` is `:eval`.
 - `always_rethrow`, if true, causes an error to be thrown if evaluating `ex` triggered an error.
   If false, the error is logged with `@error`. `InterruptException`s are always rethrown.
   This is primarily useful for debugging.
 """
-function methods_by_execution!(interp::Interpreter, methodinfo, mod::Module, ex::Expr;
-                               mode::Symbol=:eval, disablebp::Bool=true, always_rethrow::Bool=false, kwargs...)
+function methods_by_execution!(
+        interp::Interpreter, exinfo::ExInfo, mod::Module, ex::Expr;
+        mode::Symbol = :eval, disablebp::Bool = true, always_rethrow::Bool = false, kwargs...
+    )
     mode ∈ (:sigs, :eval, :evalmeth, :evalassign) || error("unsupported mode ", mode)
     lwr = Meta.lower(mod, ex)
     isa(lwr, Expr) || return Pair{Any,Union{Nothing,Expr}}(nothing, nothing)
@@ -252,11 +221,11 @@ function methods_by_execution!(interp::Interpreter, methodinfo, mod::Module, ex:
     frame = Frame(mod, lwr.args[1]::CodeInfo)
     mode === :eval || LoweredCodeUtils.rename_framemethods!(interp, frame)
     # Determine whether we need interpreted mode
-    isrequired, evalassign = minimal_evaluation!(methodinfo, frame, mode)
+    isrequired, evalassign = minimal_evaluation!(frame, mode)
     # LoweredCodeUtils.print_with_code(stdout, frame.framecode.src, isrequired)
-    if !any(isrequired) && (mode===:eval || !evalassign)
+    if !any(isrequired) && (mode === :eval || !evalassign)
         # We can evaluate the entire expression in compiled mode
-        if mode===:eval
+        if mode === :eval
             ret = try
                 Core.eval(mod, ex)
             catch err
@@ -281,9 +250,9 @@ function methods_by_execution!(interp::Interpreter, methodinfo, mod::Module, ex:
             foreach(disable, active_bp_refs)
         end
         ret = try
-            _methods_by_execution!(interp, methodinfo, frame, isrequired; mode, kwargs...)
+            _methods_by_execution!(interp, exinfo, frame, isrequired; mode, kwargs...)
         catch err
-            (always_rethrow || isa(err, InterruptException)) && (disablebp && foreach(enable, active_bp_refs); rethrow(err))
+            (always_rethrow || isa(err, InterruptException)) && (@isdefined(active_bp_refs) && foreach(enable, active_bp_refs); rethrow(err))
             loc = location_string(whereis(frame))
             sfs = []  # crafted for interaction with Base.show_backtrace
             frame = JuliaInterpreter.leaf(frame)
@@ -293,23 +262,25 @@ function methods_by_execution!(interp::Interpreter, methodinfo, mod::Module, ex:
             end
             throw(ReviseEvalException(loc, err, sfs))
         end
-        if disablebp
+        if @isdefined active_bp_refs
             foreach(enable, active_bp_refs)
         end
     end
     return Pair{Any,Union{Nothing,Expr}}(ret, lwr)
 end
-methods_by_execution!(methodinfo, mod::Module, ex::Expr; kwargs...) =
-    methods_by_execution!(Compiled(), methodinfo, mod, ex; kwargs...)
+methods_by_execution!(exinfo::ExInfo, mod::Module, ex::Expr; kwargs...) =
+    methods_by_execution!(Compiled(), exinfo, mod, ex; kwargs...)
 
-function _methods_by_execution!(interp::Interpreter, methodinfo, frame::Frame, isrequired::AbstractVector{Bool};
-                                mode::Symbol=:eval, skip_include::Bool=true)
+function _methods_by_execution!(
+        interp::Interpreter, exinfo::ExInfo, frame::Frame, isrequired::AbstractVector{Bool};
+        mode::Symbol = :eval, skip_include::Bool = true
+    )
     isok(lnn::LineTypes) = !iszero(lnn.line) || lnn.file !== :none   # might fail either one, but accept anything
 
     mod = moduleof(frame)
     # Hoist this lookup for performance. Don't throw even when `mod` is a baremodule:
-    modinclude = isdefined(mod, :include) ? getfield(mod, :include) : nothing
-    signatures = []  # temporary for method signature storage
+    modinclude = isdefined(mod, :include) ? getglobal(mod, :include) : nothing
+    signatures = MethodInfoKey[]  # temporary for method signature storage
     pc = frame.pc
     while true
         JuliaInterpreter.is_leaf(frame) || (@warn("not a leaf"); break)
@@ -325,9 +296,9 @@ function _methods_by_execution!(interp::Interpreter, methodinfo, frame::Frame, i
                 local value
                 for ex in stmt.args
                     ex isa Expr || continue
-                    value, _ = methods_by_execution!(interp, methodinfo, mod, ex; mode, disablebp=false, skip_include)
+                    value, _ = methods_by_execution!(interp, exinfo, mod, ex; mode, disablebp=false, skip_include)
                 end
-                isassign(frame, pc) && assign_this!(frame, value)
+                isassign(frame, pc) && @isdefined(value) && assign_this!(frame, value)
                 pc = next_or_nothing!(frame)
             elseif head === :thunk && defines_function(only(stmt.args))
                 mode !== :sigs && Core.eval(mod, stmt)
@@ -351,7 +322,7 @@ function _methods_by_execution!(interp::Interpreter, methodinfo, frame::Frame, i
                             file, line = loc
                             lnn = LineNumberNode(Int(line), Symbol(file))
                             for sig in signatures
-                                add_signature!(methodinfo, sig, lnn)
+                                add_signature!(exinfo, sig, lnn)
                             end
                         end
                     end
@@ -436,7 +407,7 @@ function _methods_by_execution!(interp::Interpreter, methodinfo, frame::Frame, i
                     end
                     if lnn !== nothing && isok(lnn)
                         for sig in signatures
-                            add_signature!(methodinfo, sig, lnn)
+                            add_signature!(exinfo, sig, lnn)
                         end
                     end
                 end
@@ -448,27 +419,31 @@ function _methods_by_execution!(interp::Interpreter, methodinfo, frame::Frame, i
                     pc = step_expr!(interp, frame, stmt, true)
                 end
             elseif head === :call
-                f = lookup(interp, frame, stmt.args[1])
-                if isdefined(Core, :_defaultctors) && f === Core._defaultctors && length(stmt.args) == 3
-                    T = lookup(interp, frame, stmt.args[2])
-                    lnn = lookup(interp, frame, stmt.args[3])
+                f = lookup(frame, stmt.args[1])
+                if __bpart__ && f === Core._typebody!
+                    analyze_typebody!(exinfo, interp, frame, stmt)
+                    pc = step_expr!(interp, frame, stmt, true)
+                elseif @static(isdefined(Core, :_defaultctors) && true) && f === Core._defaultctors && length(stmt.args) == 3
+                    # Create the constructors for a type (i.e., a method definition)
+                    T = lookup(frame, stmt.args[2])
+                    lnn = lookup(frame, stmt.args[3])
                     if T isa Type && lnn isa LineNumberNode
                         empty!(signatures)
                         uT = Base.unwrap_unionall(T)::DataType
                         ft = uT.types
-                        sig1 = Tuple{Base.rewrap_unionall(Type{uT}, T), Any[Any for i in 1:length(ft)]...}
-                        push!(signatures, sig1)
+                        sig1 = Tuple{Base.rewrap_unionall(Type{uT}, T), Any[Any for _ in 1:length(ft)]...}
+                        push!(signatures, MethodInfoKey(nothing, sig1))
                         sig2 = Base.rewrap_unionall(Tuple{Type{T}, ft...}, T)
                         while T isa UnionAll
                             sig2 isa UnionAll || (sig2 = sig1; break) # sig2 doesn't define all parameters, so drop it
                             T = T.body
                         end
-                        sig1 == sig2 || push!(signatures, sig2)
+                        sig1 == sig2 || push!(signatures, MethodInfoKey(nothing, sig2))
                         for sig in signatures
-                            add_signature!(methodinfo, sig, lnn)
+                            add_signature!(exinfo, sig, lnn)
                         end
                     end
-                    if mode===:sigs
+                    if mode === :sigs
                         pc = next_or_nothing!(frame)
                     else # also execute this call
                         pc = step_expr!(interp, frame, stmt, true)
@@ -477,41 +452,22 @@ function _methods_by_execution!(interp::Interpreter, methodinfo, frame::Frame, i
                     # an @eval or eval block: this may contain method definitions, so intercept it.
                     evalmod = lookup(interp, frame, stmt.args[2])::Module
                     evalex = lookup(interp, frame, stmt.args[3])
-                    value = nothing
+                    local value = nothing
                     for (newmod, newex) in ExprSplitter(evalmod, evalex)
                         if is_doc_expr(newex)
                             newex = newex.args[4]
                         end
                         newex = unwrap(newex)
-                        push_expr!(methodinfo, newmod, newex)
-                        value, _ = methods_by_execution!(interp, methodinfo, newmod, newex; mode, skip_include, disablebp=false)
-                        pop_expr!(methodinfo)
+                        push!(exinfo.exprstack, newex)
+                        value, _ = methods_by_execution!(interp, exinfo, newmod, newex; mode, skip_include, disablebp=false)
+                        pop!(exinfo.exprstack)
                     end
                     assign_this!(frame, value)
                     pc = next_or_nothing!(frame)
-                elseif skip_include && (f === modinclude || f === Core.include)
+                elseif skip_include && (f === modinclude || f === Core.include || f === Base.include)
                     # include calls need to be managed carefully from several standpoints, including
                     # path management and parsing new expressions
-                    if length(stmt.args) == 2
-                        add_includes!(methodinfo, mod, lookup(interp, frame, stmt.args[2]))
-                    elseif length(stmt.args) == 3
-                        add_includes!(methodinfo, lookup(interp, frame, stmt.args[2]), lookup(interp, frame, stmt.args[3]))
-                    else
-                        error("Bad call to Core.include")
-                    end
-                    assign_this!(frame, nothing)  # FIXME: the file might return something different from `nothing`
-                    pc = next_or_nothing!(frame)
-                elseif skip_include && f === Base.include
-                    if length(stmt.args) == 2
-                        add_includes!(methodinfo, mod, lookup(interp, frame, stmt.args[2]))
-                    else # either include(module, path) or include(mapexpr, path)
-                        mod_or_mapexpr = lookup(interp, frame, stmt.args[2])
-                        if isa(mod_or_mapexpr, Module)
-                            add_includes!(methodinfo, mod_or_mapexpr, lookup(interp, frame, stmt.args[3]))
-                        else
-                            error("include(mapexpr, path) is not supported") # TODO (issue #634)
-                        end
-                    end
+                    handle_include!(exinfo, interp, frame, stmt)
                     assign_this!(frame, nothing)  # FIXME: the file might return something different from `nothing`
                     pc = next_or_nothing!(frame)
                 elseif f === Base.Docs.doc! # && mode !== :eval
@@ -562,3 +518,48 @@ function _methods_by_execution!(interp::Interpreter, methodinfo, frame::Frame, i
     end
     return isrequired[frame.pc] ? get_return(frame) : nothing
 end
+
+function add_signature!(exinfo::ExInfo, mt_sig::MethodInfoKey, ln::LineNumberNode)
+    locdefs = CodeTracking.invoked_get!(Vector{Tuple{LineNumberNode,Expr}}, CodeTracking.method_info, mt_sig)
+    newdef = unwrap(exinfo.exprstack[end])
+    if newdef !== nothing
+        if !any(locdef->locdef[1] == ln && isequal(RelocatableExpr(locdef[2]), RelocatableExpr(newdef)), locdefs)
+            push!(locdefs, (fixpath(ln), newdef))
+        end
+        push!(exinfo.allsigs, SigInfo(mt_sig))
+    end
+    return exinfo
+end
+
+function handle_include!(exinfo::ExInfo, interp::Interpreter, frame::Frame, stmt::Expr)
+    if length(stmt.args) == 2
+        local arg2 = lookup(interp, frame, stmt.args[2])
+        if arg2 isa AbstractString
+            push!(exinfo.includes, moduleof(frame)=>arg2)
+            return exinfo
+        end
+        error("Bad include call")
+    elseif length(stmt.args) == 3
+        local arg2 = lookup(interp, frame, stmt.args[2])
+        local arg3 = lookup(interp, frame, stmt.args[3])
+        if arg2 isa Module && arg3 isa AbstractString
+            push!(exinfo.includes, arg2=>arg3)
+            return exinfo
+        end
+        error("Bad include call")
+    end
+    error("include(mapexpr::Function, mod::Module, path::AbstractString) is not supported") # TODO (issue #634)
+end
+
+function analyze_typebody!(exinfo::ExInfo, interp::Interpreter, frame::Frame, stmt::Expr)
+    if length(stmt.args) == 3 # abstract type definition
+        typ = lookup(interp, frame, stmt.args[3])::Type
+    elseif length(stmt.args) == 4 # general struct definition
+        typ = lookup(interp, frame, stmt.args[3])::Type
+    else
+        return nothing
+    end
+    datatype = Base.unwrap_unionall(typ)::DataType
+    push!(exinfo.typeinfos, TypeInfo(datatype.name))
+    return exinfo
+end
