@@ -16,23 +16,25 @@
 #     possible that this needs to be supplemented with parsing.
 function queue_includes!(pkgdata::PkgData, id::PkgId)
     modstring = id.name
-    delids = Int[]
-    for i = 1:length(included_files)
-        mod, fname = included_files[i]
-        if mod == Base.__toplevel__
-            mod = Main
-        end
-        modname = String(Symbol(mod))
-        if startswith(modname, modstring) || endswith(fname, modstring*".jl")
-            modexsigs = parse_source(fname, mod)
-            if modexsigs !== nothing
-                fname = relpath(fname, pkgdata)
-                push!(pkgdata, fname=>FileInfo(modexsigs))
+    @lock included_files_lock begin
+        delids = Int[]
+        for i = 1:length(included_files)
+            mod, fname = included_files[i]
+            if mod == Base.__toplevel__
+                mod = Main
+            end
+            modname = String(Symbol(mod))
+            if startswith(modname, modstring) || endswith(fname, modstring*".jl")
+                mod_exs_infos = parse_source(fname, mod)
+                if mod_exs_infos !== nothing
+                    fname = relpath(fname, pkgdata)
+                    push!(pkgdata, fname=>FileInfo(mod_exs_infos))
+                end
+                push!(delids, i)
             end
-            push!(delids, i)
         end
+        deleteat!(included_files, delids)
     end
-    deleteat!(included_files, delids)
     CodeTracking._pkgfiles[id] = pkgdata.info
     return pkgdata
 end
@@ -47,7 +49,7 @@ function queue_includes(mod::Module)
     if has_writable_paths(pkgdata)
         init_watching(pkgdata)
     end
-    pkgdatas[id] = pkgdata
+    @lock pkgdatas_lock pkgdatas[id] = pkgdata
     return pkgdata
 end
 
@@ -57,13 +59,15 @@ end
 function remove_from_included_files(modsym::Symbol)
     i = 1
     modstring = string(modsym)
-    while i <= length(included_files)
-        mod, fname = included_files[i]
-        modname = String(Symbol(mod))
-        if startswith(modname, modstring) || endswith(fname, modstring*".jl")
-            deleteat!(included_files, i)
-        else
-            i += 1
+    @lock included_files_lock begin
+        while i <= length(included_files)
+            mod, fname = included_files[i]
+            modname = String(Symbol(mod))
+            if startswith(modname, modstring) || endswith(fname, modstring*".jl")
+                deleteat!(included_files, i)
+            else
+                i += 1
+            end
         end
     end
 end
@@ -86,13 +90,13 @@ function maybe_parse_from_cache!(pkgdata::PkgData, file::AbstractString)
         return add_definitions_from_repl(file)
     end
     fi = fileinfo(pkgdata, file)
-    if (isempty(fi.modexsigs) && !fi.parsed[]) && (!isempty(fi.cachefile) || !isempty(fi.cacheexprs))
+    if (isempty(fi.mod_exs_infos) && !fi.parsed[]) && (!isempty(fi.cachefile) || !isempty(fi.cacheexprs))
         # Source was never parsed, get it from the precompile cache
         src = read_from_cache(pkgdata, file)
         filep = joinpath(basedir(pkgdata), file)
         filec = get(cache_file_key, filep, filep)
-        topmod = first(keys(fi.modexsigs))
-        ret = parse_source!(fi.modexsigs, src, filec, topmod)
+        topmod = first(keys(fi.mod_exs_infos))
+        ret = parse_source!(fi.mod_exs_infos, src, filec, topmod)
         if ret === nothing
             @error "failed to parse cache file source text for $file"
         end
@@ -105,26 +109,61 @@ function maybe_parse_from_cache!(pkgdata::PkgData, file::AbstractString)
     return fi
 end
 
-function add_modexs!(fi::FileInfo, modexs)
+function add_modexs!(fi::FileInfo, modexs::Vector{Tuple{Module,Expr}})
     for (mod, rex) in modexs
-        exsigs = get(fi.modexsigs, mod, nothing)
-        if exsigs === nothing
-            fi.modexsigs[mod] = exsigs = ExprsSigs()
+        exs_infos = get(fi.mod_exs_infos, mod, nothing)
+        if exs_infos === nothing
+            fi.mod_exs_infos[mod] = exs_infos = ExprsInfos()
         end
-        pushex!(exsigs, rex)
+        pushex!(exs_infos, rex)
     end
     return fi
 end
 
 function maybe_extract_sigs!(fi::FileInfo)
     if !fi.extracted[]
-        instantiate_sigs!(fi.modexsigs)
+        instantiate_sigs!(fi.mod_exs_infos)
         fi.extracted[] = true
     end
     return fi
 end
 maybe_extract_sigs!(pkgdata::PkgData, file::AbstractString) = maybe_extract_sigs!(fileinfo(pkgdata, file))
 
+is_not_populated(fi::FileInfo) =
+    (isempty(fi.mod_exs_infos) && !fi.parsed[]) && (!isempty(fi.cachefile) || !isempty(fi.cacheexprs))
+
+function maybe_extract_sigs_for_meths(meths)
+    for m in meths
+        methinfo = get(CodeTracking.method_info, MethodInfoKey(m), false)
+        if methinfo === false
+            pkgdata = get(pkgdatas, PkgId(m.module), nothing)
+            pkgdata === nothing && continue
+            for file in srcfiles(pkgdata)
+                fi = fileinfo(pkgdata, file)
+                if is_not_populated(fi)
+                    fi = maybe_parse_from_cache!(pkgdata, file)
+                    instantiate_sigs!(fi.mod_exs_infos)
+                end
+            end
+        end
+    end
+end
+
+function maybe_extract_sigs_for_types(types)
+    for ty in types
+        m = parentmodule(ty)
+        pkgdata = get(pkgdatas, PkgId(m), nothing)
+        pkgdata === nothing && continue
+        for file in srcfiles(pkgdata)
+            fi = fileinfo(pkgdata, file)
+            if is_not_populated(fi)
+                fi = maybe_parse_from_cache!(pkgdata, file)
+                instantiate_sigs!(fi.mod_exs_infos)
+            end
+        end
+    end
+end
+
 function maybe_add_includes_to_pkgdata!(pkgdata::PkgData, file::AbstractString, includes; eval_now::Bool=false)
     for (mod, inc) in includes
         inc = joinpath(splitdir(file)[1], inc)
@@ -144,10 +183,10 @@ function maybe_add_includes_to_pkgdata!(pkgdata::PkgData, file::AbstractString,
             # Parse the source of the new file
             fullfile = joinpath(basedir(pkgdata), incrp)
             if isfile(fullfile)
-                parse_source!(fi.modexsigs, fullfile, mod)
+                parse_source!(fi.mod_exs_infos, fullfile, mod)
                 if eval_now
                     # Use runtime dispatch to reduce latency
-                    Base.invokelatest(instantiate_sigs!, fi.modexsigs; mode=:eval)
+                    Base.invokelatest(instantiate_sigs!, fi.mod_exs_infos; mode=:eval)
                 end
             end
             # Add to watchlist
@@ -160,7 +199,9 @@ end
 # Use locking to prevent races between inner and outer @require blocks
 const requires_lock = ReentrantLock()
 
-function add_require(sourcefile::String, modcaller::Module, idmod::String, modname::String, expr::Expr)
+# This is used by Requires.jl: therefore even if it appears unused by Revise.jl,
+# it cannot be removed as long as we support integration with Requires.jl
+function add_require(sourcefile::String, modcaller::Module, idmod::String, ::String, expr::Expr)
     id = PkgId(modcaller)
     # If this fires when the module is first being loaded (because the dependency
     # was already loaded), Revise may not yet have the pkgdata for this package.
@@ -168,8 +209,7 @@ function add_require(sourcefile::String, modcaller::Module, idmod::String, modna
         watch_package(id)
     end
 
-    lock(requires_lock)
-    try
+    @lock requires_lock begin
         # Get/create the FileInfo specifically for tracking @require blocks
         pkgdata = pkgdatas[id]
         filekey = relpath(sourcefile, pkgdata) * "__@require__"
@@ -197,23 +237,21 @@ function add_require(sourcefile::String, modcaller::Module, idmod::String, modna
                     push!(modincludes, (modcaller, inc))
                 end
                 maybe_add_includes_to_pkgdata!(pkgdata, filekey, modincludes)
-                if isempty(fi.modexsigs)
+                if isempty(fi.mod_exs_infos)
                     # Source has not even been parsed
                     push!(fi.cacheexprs, (modcaller, expr))
                 else
-                    add_modexs!(fi, [(modcaller, expr)])
+                    add_modexs!(fi, Tuple{Module,Expr}[(modcaller, expr)])
                 end
             end
         end
         if complex
             Base.invokelatest(eval_require_now, pkgdata, fileidx, filekey, sourcefile, modcaller, expr)
         end
-    finally
-        unlock(requires_lock)
     end
 end
 
-function deferrable_require(expr)
+function deferrable_require(expr::Expr)
     includes = String[]
     complex = deferrable_require!(includes, expr)
     return includes, complex
@@ -242,16 +280,16 @@ end
 
 function eval_require_now(pkgdata::PkgData, fileidx::Int, filekey::String, sourcefile::String, modcaller::Module, expr::Expr)
     fi = pkgdata.fileinfos[fileidx]
-    exsnew = ExprsSigs()
-    exsnew[RelocatableExpr(expr)] = nothing
-    mexsnew = ModuleExprsSigs(modcaller=>exsnew)
+    exs_infos_new = ExprsInfos()
+    exs_infos_new[RelocatableExpr(expr)] = nothing
+    mod_exs_infos_new = ModuleExprsInfos(modcaller=>exs_infos_new)
     # Before executing the expression we need to set the load path appropriately
     prev = Base.source_path(nothing)
     tls = task_local_storage()
     tls[:SOURCE_PATH] = sourcefile
     # Now execute the expression
-    mexsnew, includes = try
-        eval_new!(mexsnew, fi.modexsigs)
+    mod_exs_infos_new, includes = try
+        eval_new!(mod_exs_infos_new, fi.mod_exs_infos)
     finally
         if prev === nothing
             delete!(tls, :SOURCE_PATH)
@@ -260,12 +298,12 @@ function eval_require_now(pkgdata::PkgData, fileidx::Int, filekey::String, sourc
         end
     end
     # Add any new methods or `include`d files to tracked objects
-    pkgdata.fileinfos[fileidx] = FileInfo(mexsnew, fi)
+    pkgdata.fileinfos[fileidx] = FileInfo(mod_exs_infos_new, fi)
     ret = maybe_add_includes_to_pkgdata!(pkgdata, filekey, includes; eval_now=true)
     return ret
 end
 
-function watch_files_via_dir(dirname)
+function watch_files_via_dir(dirname::AbstractString)
     try
         wait_changed(dirname)  # this will block until there is a modification
     catch e
@@ -316,11 +354,10 @@ function watch_package(id::PkgId)
 
     pkgdata = get(pkgdatas, id, nothing)
     pkgdata !== nothing && return pkgdata
-    lock(wplock)
-    try
+    @lock wplock begin
         modsym = Symbol(id.name)
         if modsym ∈ dont_watch_pkgs
-            if modsym ∉ silence_pkgs
+            if id.name ∉ silence_pkgs
                 @warn "$modsym is excluded from watching by Revise. Use Revise.silence(\"$modsym\") to quiet this warning."
             end
             remove_from_included_files(modsym)
@@ -330,9 +367,7 @@ function watch_package(id::PkgId)
         if has_writable_paths(pkgdata)
             init_watching(pkgdata, srcfiles(pkgdata))
         end
-        pkgdatas[id] = pkgdata
-    finally
-        unlock(wplock)
+        @lock pkgdatas_lock pkgdatas[id] = pkgdata
     end
     return pkgdata
 end
@@ -359,7 +394,7 @@ function has_writable_paths(pkgdata::PkgData)
 end
 
 function watch_includes(mod::Module, fn::AbstractString)
-    push!(included_files, (mod, normpath(abspath(fn))))
+    @lock included_files_lock push!(included_files, (mod, abspath_no_normalize(fn)))
 end
 
 ## Working with Pkg and code-loading
@@ -416,55 +451,23 @@ function watch_manifest(mfile::String)
                 @debug "Pkg" _group="manifest_update" manifest_file=mfile
                 isfile(mfile) || return nothing
                 pkgdirs = manifest_paths(mfile)
+                pathreplacements = Pair{String,String}[]
                 for (id, pkgdir) in pkgdirs
                     if haskey(pkgdatas, id)
                         pkgdata = pkgdatas[id]
-                        if pkgdir != basedir(pkgdata)
+                        if !samefile(pkgdir, basedir(pkgdata))
                             ## The package directory has changed
                             @debug "Pkg" _group="pathswitch" oldpath=basedir(pkgdata) newpath=pkgdir
-                            # Stop all associated watching tasks
-                            for dir in unique_dirs(srcfiles(pkgdata))
-                                @debug "Pkg" _group="unwatch" dir=dir
-                                delete!(watched_files, joinpath(basedir(pkgdata), dir))
-                                # Note: if the file is revised, the task(s) will run one more time.
-                                # However, because we've removed the directory from the watch list this will be a no-op,
-                                # and then the tasks will be dropped.
-                            end
-                            # Revise code as needed
-                            files = String[]
-                            mustnotify = false
-                            for file in srcfiles(pkgdata)
-                                fi = try
-                                    maybe_parse_from_cache!(pkgdata, file)
-                                catch err
-                                    # https://github.com/JuliaLang/julia/issues/42404
-                                    # Get the source-text from the package source instead
-                                    fi = fileinfo(pkgdata, file)
-                                    if isempty(fi.modexsigs) && (!isempty(fi.cachefile) || !isempty(fi.cacheexprs))
-                                        filep = joinpath(basedir(pkgdata), file)
-                                        src = read(filep, String)
-                                        topmod = first(keys(fi.modexsigs))
-                                        if parse_source!(fi.modexsigs, src, filep, topmod) === nothing
-                                            @error "failed to parse source text for $filep"
-                                        end
-                                        add_modexs!(fi, fi.cacheexprs)
-                                        empty!(fi.cacheexprs)
-                                        fi.parsed[] = true
-                                    end
-                                    fi
-                                end
-                                maybe_extract_sigs!(fi)
-                                push!(revision_queue, (pkgdata, file))
-                                push!(files, file)
-                                mustnotify = true
-                            end
-                            mustnotify && notify(revision_event)
-                            # Update the directory
-                            pkgdata.info.basedir = pkgdir
-                            # Restart watching, if applicable
-                            if has_writable_paths(pkgdata)
-                                init_watching(pkgdata, files)
-                            end
+                            push!(pathreplacements, basedir(pkgdata)=>pkgdir)
+                            switch_basepath(pkgdata, pkgdir)
+                        end
+                    end
+                end
+                # Update the paths in the watchlist
+                for (oldpath, newpath) in pathreplacements
+                    for (_, pkgdata) in pkgdatas
+                        if samefile(basedir(pkgdata), oldpath)
+                            switch_basepath(pkgdata, newpath)
                         end
                     end
                 end
@@ -475,6 +478,53 @@ function watch_manifest(mfile::String)
     end
 end
 
+function switch_basepath(pkgdata::PkgData, newpath::String)
+    # Stop all associated watching tasks
+    for dir in unique_dirs(srcfiles(pkgdata))
+        @debug "Pkg" _group="unwatch" dir=dir
+        @lock watched_files_lock delete!(watched_files, joinpath(basedir(pkgdata), dir))
+        # Note: if the file is revised, the task(s) will run one more time.
+        # However, because we've removed the directory from the watch list this will be a no-op,
+        # and then the tasks will be dropped.
+    end
+    # Revise code as needed
+    files = String[]
+    mustnotify = false
+    for file in srcfiles(pkgdata)
+        fi = try
+            maybe_parse_from_cache!(pkgdata, file)
+        catch
+            # https://github.com/JuliaLang/julia/issues/42404
+            # Get the source-text from the package source instead
+            fi = fileinfo(pkgdata, file)
+            if isempty(fi.mod_exs_infos) && (!isempty(fi.cachefile) || !isempty(fi.cacheexprs))
+                filep = joinpath(basedir(pkgdata), file)
+                src = read(filep, String)
+                topmod = first(keys(fi.mod_exs_infos))
+                if parse_source!(fi.mod_exs_infos, src, filep, topmod) === nothing
+                    @error "failed to parse source text for $filep"
+                end
+                add_modexs!(fi, fi.cacheexprs)
+                empty!(fi.cacheexprs)
+                fi.parsed[] = true
+            end
+            fi
+        end
+        maybe_extract_sigs!(fi)
+        push!(revision_queue, (pkgdata, file))
+        push!(files, file)
+        mustnotify = true
+    end
+    mustnotify && notify(revision_event)
+    # Update the directory
+    pkgdata.info.basedir = newpath
+    # Restart watching, if applicable
+    if has_writable_paths(pkgdata)
+        init_watching(pkgdata, files)
+    end
+    return nothing
+end
+
 function active_project_watcher()
     mfile = manifest_file()
     if !isnothing(mfile) && mfile ∉ watched_manifests
