@@ -31,15 +31,15 @@ function _precompile_()
     # setindex! doesn't fully precompile, but it's still beneficial to do it
     # (it shaves off a bit of the time)
     # See https://github.com/JuliaLang/julia/pull/31466
-    @warnpcfail precompile(Tuple{typeof(setindex!), ExprsSigs, Nothing, RelocatableExpr})
-    @warnpcfail precompile(Tuple{typeof(setindex!), ExprsSigs, Vector{Any}, RelocatableExpr})
-    @warnpcfail precompile(Tuple{typeof(setindex!), ModuleExprsSigs, ExprsSigs, Module})
+    @warnpcfail precompile(Tuple{typeof(setindex!), ExprsInfos, Nothing, RelocatableExpr})
+    @warnpcfail precompile(Tuple{typeof(setindex!), ExprsInfos, Vector{Any}, RelocatableExpr})
+    @warnpcfail precompile(Tuple{typeof(setindex!), ModuleExprsInfos, ExprsInfos, Module})
     @warnpcfail precompile(Tuple{typeof(setindex!), Dict{PkgId,PkgData}, PkgData, PkgId})
     @warnpcfail precompile(Tuple{Type{WatchList}})
     @warnpcfail precompile(Tuple{typeof(setindex!), Dict{String,WatchList}, WatchList, String})
 
-    MI = CodeTrackingMethodInfo
-    @warnpcfail precompile(Tuple{typeof(minimal_evaluation!), Any, MI, Module, Core.CodeInfo, Symbol})
+    MI = ExInfo
+    @warnpcfail precompile(Tuple{typeof(minimal_evaluation!), Any, Module, Core.CodeInfo, Symbol})
     @warnpcfail precompile(Tuple{typeof(methods_by_execution!), Compiled, MI, Module, Expr})
     @warnpcfail precompile(Tuple{typeof(_methods_by_execution!), Compiled, MI, Frame, Vector{Bool}})
     @warnpcfail precompile(Tuple{typeof(Core.kwfunc(methods_by_execution!)),
@@ -62,12 +62,11 @@ function _precompile_()
     mbody = bodymethod(mex)
     # use `typeof(pairs(NamedTuple()))` here since it actually differs between Julia versions
     @warnpcfail precompile(Tuple{mbody.sig.parameters[1], Symbol, Bool, Bool, typeof(pairs(NamedTuple())), typeof(methods_by_execution!), Compiled, MI, Module, Expr})
-    @warnpcfail precompile(Tuple{mbody.sig.parameters[1], Symbol, Bool, Bool, Iterators.Pairs{Symbol,Bool,Tuple{Symbol},NamedTuple{(:skip_include,),Tuple{Bool}}}, typeof(methods_by_execution!), Compiled, MI, Module, Expr})
+    @warnpcfail precompile(Tuple{mbody.sig.parameters[1], Symbol, Bool, Bool, Iterators.Pairs{Symbol,Bool,Nothing,NamedTuple{(:skip_include,),Tuple{Bool}}}, typeof(methods_by_execution!), Compiled, MI, Module, Expr})
     mfr = which(_methods_by_execution!, (Compiled, MI, Frame, Vector{Bool}))
     mbody = bodymethod(mfr)
     @warnpcfail precompile(Tuple{mbody.sig.parameters[1], Symbol, Bool, typeof(_methods_by_execution!), Compiled, MI, Frame, Vector{Bool}})
 
-    @warnpcfail precompile(Tuple{typeof(hastrackedexpr), Expr, Vector{Any}})
     @warnpcfail precompile(Tuple{typeof(get_def), Method})
     @warnpcfail precompile(Tuple{typeof(parse_pkg_files), PkgId})
     if isdefined(Revise, :filter_valid_cachefiles)
@@ -77,8 +76,8 @@ function _precompile_()
     @warnpcfail precompile(Tuple{typeof(Revise.active_repl_backend_available)})
     @warnpcfail precompile(Tuple{typeof(pkg_fileinfo), PkgId})
     @warnpcfail precompile(Tuple{typeof(push!), WatchList, Pair{String,PkgId}})
-    @warnpcfail precompile(Tuple{typeof(pushex!), ExprsSigs, Expr})
-    @warnpcfail precompile(Tuple{Type{ModuleExprsSigs}, Module})
+    @warnpcfail precompile(Tuple{typeof(pushex!), ExprsInfos, Expr})
+    @warnpcfail precompile(Tuple{Type{ModuleExprsInfos}, Module})
     @warnpcfail precompile(Tuple{Type{FileInfo}, Module, String})
     @warnpcfail precompile(Tuple{Type{PkgData}, PkgId})
     @warnpcfail precompile(Tuple{typeof(Base._deleteat!), Vector{Tuple{Module,String,Float64}}, Vector{Int}})
