@@ -1,17 +1,23 @@
 """
-    Revise.track(Base)
-    Revise.track(Core.Compiler)
-    Revise.track(stdlib)
+    Revise.track(Base; revise_throw::Bool=!isinteractive())
+    Revise.track(Core.Compiler; revise_throw::Bool=!isinteractive())
+    Revise.track(stdlib; revise_throw::Bool=!isinteractive())
 
 Track updates to the code in Julia's `base` directory, `base/compiler`, or one of its
-standard libraries.
+standard libraries. Calls `revise()` after tracking to ensure that any changes
+detected during tracking are applied immediately. Optionally, if `revise_throw` is
+`true`, `revise()` will throw if any exceptions are encountered while revising.
 """
-function track(mod::Module; modified_files=revision_queue)
-    id = PkgId(mod)
+function track(mod::Module; modified_files=revision_queue, revise_throw::Bool=!isinteractive())
+    id = pkgidid_for_mod(mod)
     modname = nameof(mod)
-    return _track(id, modname; modified_files=modified_files)
+    ret = _track(id, modname; modified_files=modified_files)
+    revise(; throw=revise_throw) # force revision so following calls in the same block work
+    return ret
 end
 
+pkgidid_for_mod(mod) = id = Base.moduleroot(mod) == Core.Compiler ? PkgId(mod, "Core.Compiler") : PkgId(mod)
+
 const vstring = "v$(VERSION.major).$(VERSION.minor)"
 
 function inpath(path::AbstractString, dirs::Vector{String})
@@ -61,8 +67,17 @@ function _track(id::PkgId, modname::Symbol; modified_files=revision_queue)
         if pkgdata === nothing
             pkgdata = PkgData(id, srcdir)
         end
-        lock(revise_lock) do
-            for (submod, filename) in Iterators.drop(Base._included_files, 1)  # stepping through sysimg.jl rebuilds Base, omit it
+        ret = Revise.pkg_fileinfo(id)
+        if ret !== nothing
+            cachefile, _ = ret
+            if cachefile === nothing
+                @error "unable to find cache file for $id, tracking is not possible"
+            end
+        else
+            cachefile = basesrccache
+        end
+        @lock revision_queue_lock begin
+            for (submod, filename) in modulefiles_basestlibs(id)
                 ffilename = fixpath(filename)
                 inpath(ffilename, dirs) || continue
                 keypath = ffilename[1:last(findfirst(dirs[end], ffilename))]
@@ -72,7 +87,7 @@ function _track(id::PkgId, modname::Symbol; modified_files=revision_queue)
                     cache_file_key[fullpath] = filename
                     src_file_key[filename] = fullpath
                 end
-                push!(pkgdata, rpath=>FileInfo(submod, basesrccache))
+                push!(pkgdata, rpath=>FileInfo(submod, cachefile))
                 if mtime(ffilename) > mtcache
                     with_logger(_debug_logger) do
                         @debug "Recipe for Base/StdLib" _group="Watching" filename=filename mtime=mtime(filename) mtimeref=mtcache
@@ -86,10 +101,10 @@ function _track(id::PkgId, modname::Symbol; modified_files=revision_queue)
         # Add the files to the watch list
         init_watching(pkgdata, srcfiles(pkgdata))
         # Save the result (unnecessary if already in pkgdatas, but doesn't hurt either)
-        pkgdatas[id] = pkgdata
+        @lock pkgdatas_lock pkgdatas[id] = pkgdata
     elseif modname === :Compiler
-        compilerdir = normpath(joinpath(juliadir, "Compiler", "src"))
-        compilerdir_pre_112 = normpath(joinpath(juliadir, "base", "compiler"))
+        compilerdir = joinpath(juliadir, "Compiler", "src")
+        compilerdir_pre_112 = joinpath(juliadir, "base", "compiler")
         isdir(compilerdir) || (compilerdir = compilerdir_pre_112)
         pkgdata = get(pkgdatas, id, nothing)
         if pkgdata === nothing
@@ -106,7 +121,6 @@ end
 # Fix paths to files that define Julia (base and stdlibs)
 function fixpath(filename::AbstractString; badpath=basebuilddir, goodpath=juliadir)
     startswith(filename, badpath) || return normpath(filename)
-    filec = filename
     relfilename = relpath(filename, badpath)
     relfilename0 = relfilename
     for strippath in (#joinpath("usr", "share", "julia", "stdlib", "v$(VERSION.major).$(VERSION.minor)"),
@@ -135,11 +149,11 @@ function track_subdir_from_git!(pkgdata::PkgData, subdir::AbstractString; commit
     if repo == nothing
         throw(GitRepoException(subdir))
     end
-    prefix = string(relpath(subdir, repo_path), "/")   # git-relative path of this subdir
+    prefix = string(relpath(realpath(subdir), realpath(repo_path)), "/")   # git-relative path of this subdir
     tree = git_tree(repo, commit)
     files = Iterators.filter(file->startswith(file, prefix) && endswith(file, ".jl"), keys(tree))
     ccall((:giterr_clear, :libgit2), Cvoid, ())  # necessary to avoid errors like "the global/xdg file 'attributes' doesn't exist: No such file or directory"
-    lock(revise_lock) do
+    @lock revision_queue_lock begin
         for file in files
             fullpath = joinpath(repo_path, file)
             rpath = relpath(fullpath, pkgdata)  # this might undo the above, except for Core.Compiler
@@ -155,22 +169,28 @@ function track_subdir_from_git!(pkgdata::PkgData, subdir::AbstractString; commit
             end
             fmod = get(juliaf2m, fullpath, Core.Compiler)  # Core.Compiler is not cached
             if fmod === Core.Compiler
-                endswith(fullpath, "compiler.jl") && continue  # defines the module, skip
+                endswith(fullpath, "compiler.jl") && continue  # defines the module (v1.11-), skip
+                endswith(fullpath, "/Compiler/src/Compiler.jl") && continue  # defines the module (v1.12+), skip
                 @static if isdefined(Core.Compiler, :EscapeAnalysis)
                     # after https://github.com/JuliaLang/julia/pull/43800
-                    if contains(fullpath, "compiler/ssair/EscapeAnalysis")
+                    if endswith(fullpath, "/compiler/ssair/EscapeAnalysis.jl") || contains(fullpath, "/Compiler/src/ssair/EscapeAnalysis.jl")
                         fmod = Core.Compiler.EscapeAnalysis
                     end
                 end
+                @static if isdefined(Core.Compiler, :TrimVerifier)
+                    if endswith(fullpath, "/Compiler/src/verifytrim.jl")
+                        fmod = Core.Compiler.TrimVerifier
+                    end
+                end
             end
             if src != read(fullpath, String)
                 push!(modified_files, (pkgdata, rpath))
             end
             fi = FileInfo(fmod)
-            if parse_source!(fi.modexsigs, src, file, fmod) === nothing
+            if parse_source!(fi.mod_exs_infos, src, file, fmod) === nothing
                 @warn "failed to parse Git source text for $file"
             else
-                instantiate_sigs!(fi.modexsigs)
+                instantiate_sigs!(fi.mod_exs_infos)
             end
             push!(pkgdata, rpath=>fi)
         end
@@ -179,7 +199,7 @@ function track_subdir_from_git!(pkgdata::PkgData, subdir::AbstractString; commit
         id = PkgId(pkgdata)
         CodeTracking._pkgfiles[id] = pkgdata.info
         init_watching(pkgdata, srcfiles(pkgdata))
-        pkgdatas[id] = pkgdata
+        @lock pkgdatas_lock pkgdatas[id] = pkgdata
     end
     return nothing
 end
@@ -193,11 +213,12 @@ const stdlib_names = Set([
     :Serialization, :SHA, :SharedArrays, :Sockets, :SparseArrays,
     :Statistics, :SuiteSparse, :Test, :Unicode, :UUIDs,
     :TOML, :Artifacts, :LibCURL_jll, :LibCURL, :MozillaCACerts_jll,
-    :Downloads, :Tar, :ArgTools, :NetworkOptions])
+    :Downloads, :Tar, :ArgTools, :NetworkOptions, :PCRE2_jll, :Zlib_jll])
 
 # This replacement is needed because the path written during compilation differs from
 # the git source path
-const stdlib_rep = joinpath("usr", "share", "julia", "stdlib", "v$(VERSION.major).$(VERSION.minor)") => "stdlib"
+const stdpath_rep = (joinpath("usr", "share", "julia", "stdlib", "v$(VERSION.major).$(VERSION.minor)") => "stdlib",
+                    joinpath("usr", "share", "julia", "Compiler") => "Compiler")
 
-const juliaf2m = Dict(normpath(replace(file, stdlib_rep))=>mod
+const juliaf2m = Dict(normpath(replace(file, stdpath_rep...))=>mod
     for (mod,file) in Base._included_files)
