@@ -1,18 +1,15 @@
-relpath_safe(path::AbstractString, startpath::AbstractString) = isempty(startpath) ? path : relpath(path, startpath)
+function relpath_safe(path::AbstractString, startpath::AbstractString)
+    isempty(startpath) && return path
+    if ispath(path)
+        path = relpath(realpath(path), realpath(startpath))
+        !isabspath(path) && return path
+    end
+    return relpath(path, startpath)
+end
 
 function Base.relpath(filename::AbstractString, pkgdata::PkgData)
     if isabspath(filename)
-        # `Base.locate_package`, which is how `pkgdata` gets initialized, might strip pieces of the path.
-        # For example, on Travis macOS the paths returned by `abspath`
-        # can be preceded by "/private" which is not present in the value returned by `Base.locate_package`.
-        idx = findfirst(basedir(pkgdata), filename)
-        if idx !== nothing
-            idx = first(idx)
-            if idx > 1
-                filename = filename[idx:end]
-            end
-            filename = relpath_safe(filename, basedir(pkgdata))
-        end
+        filename = relpath_safe(filename, basedir(pkgdata))
     elseif startswith(filename, "compiler")
         # Core.Compiler's pkgid includes "compiler/" in the path
         filename = relpath(filename, "compiler")
@@ -33,6 +30,20 @@ function unique_dirs(iter)
     return udirs
 end
 
+function abspath_no_normalize(a)
+    if !isabspath(a)
+        cwd = pwd()
+        a_drive, a_nodrive = splitdrive(a)
+        if a_drive != "" && lowercase(splitdrive(cwd)[1]) != lowercase(a_drive)
+            cwd = a_drive * path_separator
+            a = joinpath(cwd, a_nodrive)
+        else
+            a = joinpath(cwd, a)
+        end
+    end
+    return a
+end
+
 function file_exists(filename::AbstractString)
     filename = normpath(filename)
     isfile(filename) && return true
@@ -57,8 +68,6 @@ function firstline(ex::Expr)
 end
 firstline(rex::RelocatableExpr) = firstline(rex.ex)
 
-newloc(methloc::LineNumberNode, ln, lno) = fixpath(ln)
-
 location_string((file, line)::Tuple{AbstractString, Any},) = abspath(file)*':'*string(line)
 location_string((file, line)::Tuple{Symbol, Any},) = location_string((string(file), line))
 location_string(::Nothing) = "unknown location"
@@ -96,14 +105,14 @@ function unwrap_where(ex::Expr)
     return ex::Expr
 end
 
-function pushex!(exsigs::ExprsSigs, ex::Expr)
+function pushex!(exs_infos::ExprsInfos, ex::Expr)
     uex = unwrap(ex)
     if is_doc_expr(uex)
         body = uex.args[4]
         # Don't trigger for exprs where the documented expression is just a signature
         # (e.g. `"docstr" f(x::Int)`, `"docstr" f(x::T) where T` etc.)
         if isa(body, Expr) && unwrap_where(body).head !== :call
-            exsigs[RelocatableExpr(body)] = nothing
+            exs_infos[RelocatableExpr(body)] = nothing
         end
         if length(uex.args) < 5
             push!(uex.args, false)
@@ -111,8 +120,8 @@ function pushex!(exsigs::ExprsSigs, ex::Expr)
             uex.args[5] = false
         end
     end
-    exsigs[RelocatableExpr(ex)] = nothing
-    return exsigs
+    exs_infos[RelocatableExpr(ex)] = nothing
+    return exs_infos
 end
 
 ## WatchList utilities
@@ -178,7 +187,7 @@ function printf_maxsize(f::Function, io::IO, args...; maxchars::Integer=500, max
     for i = 1:half
         f(io, lines[i])
     end
-    maxlines > 1 && f(io, ⋮)
+    maxlines > 1 && f(io, '⋮')
     for i = length(lines) - (maxlines-half) + 1:length(lines)
         f(io, lines[i])
     end
